#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
.. module:: TODO
   :platform: Unix
   :synopsis: TODO.

.. moduleauthor:: Aljosha Friemann aljosha.friemann@gmail.com

"""

import click, logging, time, os

from . import __version__
from .consul import Consul, Service, ConsulException
from .nginx import Nginx, NginxException

logger = logging.getLogger(__name__)

def setup_logging(verbose, debug):
    if verbose or debug:
        logging.basicConfig(level=logging.DEBUG if debug else logging.INFO)
        logging.getLogger('simple_model').setLevel(logging.INFO)
        logging.getLogger('requests').setLevel(logging.INFO if debug else logging.WARNING)
    else:
        logging.basicConfig(level=logging.WARNING)

@click.command()
@click.version_option(version=__version__)
@click.option('--test/--no-test', default=True, help='test the generated file before reloading nginx')
@click.option('--reload/--no-reload', default=True, help='reload nginx after file generation')
@click.option('-v', '--verbose/--no-verbose', default=False, help='print verbose log output')
@click.option('--debug/--no-debug', default=False, help='print debug log output')
@click.option('-d', '--daemonize/--no-daemonize', default=False, help='run forever')
@click.option('-o', '--output', default='/etc/nginx/nginx.conf')
@click.option('-h', '--host', default='127.0.0.1', help='consul host')
@click.option('-p', '--port', default=80, type=int, help='consul port')
@click.option('-t', '--timeout', default=5, type=int, help='timeout for polling (only used when running as daemon)')
@click.option('--overview/--no-overview', default=True, help='create overview file')
@click.option('--nginx-root', default=os.path.expanduser('~nginx'))
def main(test, reload, verbose, debug, daemonize, output, host, port, timeout, overview, nginx_root):
    try:
        setup_logging(verbose, debug)

        logger.debug('creating consul client')

        consul = Consul(host, port)

        logger.debug('checking consul connection')

        if daemonize:
            logger.info('checking %s:%s every %ss' % (host, port, timeout))

        while True:
            try:
                logger.debug('creating nginx config from consul')
                services = list(consul.get_services())
                config = Nginx.create_config(output, services)

                logger.debug('checking for differences with current file')

                if Nginx.different(config, output):
                    logger.info('config file has changed')

                    if test:
                        logger.info('testing new config')
                        Nginx.test_config(config)

                    logger.info('writing new config to %s' % output)
                    Nginx.update_config(config, output)

                    if overview:
                        logger.info('creating new overview file at %s' % nginx_root)
                        Nginx.update_overview(Nginx.create_overview(services), nginx_root)

                    if reload:
                        logger.info('reloading nginx')
                        Nginx.reload()
            except ConsulException as e:
                logger.error('failed to connect to consul API via %s:%s: %s', host, port, e)
            except NginxException as e:
                logger.error(e)
            finally:
                if daemonize:
                    time.sleep(timeout)
                else:
                    break

        return 0
    except Exception as e:
        logger.exception(e)
        return 1

if __name__ == '__main__':
    import sys
    sys.exit(main())

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4 fenc=utf-8



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
.. module:: TODO
   :platform: Unix
   :synopsis: TODO.

.. moduleauthor:: Aljosha Friemann aljosha.friemann@gmail.com

"""

import tempfile, logging, subprocess, os

from jinja2 import Template
from jinja2 import Environment, FileSystemLoader

logger = logging.getLogger(__name__)

class NginxException(Exception):
    pass

class Nginx:
    @staticmethod
    def get_template(name):
        cwd = os.path.dirname(os.path.abspath(__file__))
        j2_env = Environment(loader=FileSystemLoader(os.path.join(cwd, 'templates')))
        return j2_env.get_template(name)

    @staticmethod
    def different(config, path):
        if not os.path.isfile(path):
            return True

        with open(path, 'r') as f:
            content = f.read()
        return config != content

    @staticmethod
    def create_config(path, services):
        config = Nginx.get_template('nginx.conf.jinja2').render(service_groups=Nginx.group_services(services))

        if config is None:
            raise NginxException('config was None after creation')

        return config

    @staticmethod
    def create_overview(services):
        overview = Nginx.get_template('index.html.jinja2').render(service_groups=Nginx.group_services(services))

        if overview is None:
            raise NginxException('overview was None after creation')

        return overview

    @staticmethod
    def update_config(config, path):
        with open(path, 'wb') as result:
            result.write(config.encode())

    @staticmethod
    def update_overview(overview, nginx_root):
        html_dir = os.path.join(nginx_root, 'html')
        if not os.path.isdir(html_dir): os.mkdir(html_dir)

        with open(os.path.join(html_dir, 'index.html'), 'wb') as result:
            result.write(overview.encode())

    @staticmethod
    def test_config(config):
        with tempfile.NamedTemporaryFile() as tmp:
            tmp.write(config.encode())
            tmp.flush()
            Nginx.check_file(tmp.name)

    @staticmethod
    def group_services(services):
        grouped_services = {'stream': {}, 'http': []}

        for service in services:
            if 'http' in service.tags:
                grouped_services['http'].append(service)
            else:
                while service.port in grouped_services['stream']:
                    service.port += 1

                if service.port > 65535:
                    logging.warning('failed to find available port for service: %s' % service)
                    continue

                grouped_services['stream'][service.port] = service

        return grouped_services

    @staticmethod
    def check_file(path):
        try:
            subprocess.check_call([os.environ.get('NGINX_BINARY', '/sbin/nginx'), '-t', '-c', path])
        except subprocess.CalledProcessError:
            raise NginxException('config file did not pass test')

    @staticmethod
    def reload():
        try:
            subprocess.check_call(['service', 'nginx', 'reload'])
        except subprocess.CalledProcessError:
            raise NginxException('failed to reload nginx')

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4 fenc=utf-8



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
.. module:: TODO
   :platform: Unix
   :synopsis: TODO.

.. moduleauthor:: Aljosha Friemann aljosha.friemann@gmail.com

"""

import requests, simple_model, json, logging

logger = logging.getLogger(__name__)

class ConsulException(Exception):
    pass

class ConsulV1:
    def __init__(self, host, port):
        self._host = host
        self._port = port

    def _get_(self, endpoint):
        try:
            return requests.get('http://%s:%s/v1/%s' % (self._host, self._port, endpoint))
        except ValueError as e:
            logger.warning('failed to parse response data, is this really a consul server?')
            raise ConsulException(endpoint, e)
        except requests.exceptions.ConnectionError as e:
            logger.error(e)
            raise ConsulException(endpoint, e)

    def get_services(self):
        services = []
        for name, tags in self._get_('catalog/services').json().items():
            try:
                port = int(name.split('-')[-1])
            except ValueError:
                logger.info('ignoring service %s due to unavailable port information' % name)
                continue

            instances = [ dict(i) for i in self.get_service_instances(name) ] # necessary due to simple_model right now
            services.append(Service(name=name, port=port, tags=tags, instances=instances))

        return services

    def get_service_instances(self, name):
        instances = []
        for entry in self._get_('catalog/service/%s' % name).json():
            instances.append(ServiceInstance(**entry))
        return instances

class ServiceInstance(simple_model.Model):
    """
    {
        "Node": "ip-10-0-1-81.eu-west-1.compute.internal",
        "Address": "10.0.1.81",
        "ServiceID": "consul",
        "ServiceName": "consul",
        "ServiceTags": [],
        "ServiceAddress": "",
        "ServicePort": 8300,
        "ServiceEnableTagOverride": false,
        "CreateIndex": 4,
        "ModifyIndex": 5
    }
    """
    Node = simple_model.Attribute(str)
    Address = simple_model.Attribute(str)
    ServiceID = simple_model.Attribute(str)
    ServiceName = simple_model.Attribute(str)
    ServiceTags = simple_model.AttributeList(str)
    ServicePort = simple_model.Attribute(int)

class Service(simple_model.Model):
    """
    {
        "consul": [],
        "consul-loadbalancer-5601": ['udp'],
        "env-loop-1200": [
            "udp"
        ],
        "env-loop-4000": [],
        "metrics-elasticsearch": [],
        "metrics-kibana": [],
        "metrics-logstash-in": [],
        "metrics-rabbitmq": []
    }
    """
    name = simple_model.Attribute(str)
    port = simple_model.Attribute(int)
    tags = simple_model.AttributeList(str)
    instances = simple_model.AttributeList(ServiceInstance)

Consul = ConsulV1

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4 fenc=utf-8



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
.. module:: TODO
   :platform: Unix
   :synopsis: TODO.

.. moduleauthor:: Aljosha Friemann aljosha.friemann@gmail.com

"""

import sys, logging

from .cli import main

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    sys.exit(main())

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4 fenc=utf-8



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
.. module:: TODO
   :platform: Unix
   :synopsis: TODO.

.. moduleauthor:: Aljosha Friemann aljosha.friemann@gmail.com

"""

import os, pip

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname), 'r').read()

install_reqs = pip.req.parse_requirements('requirements.txt', session=pip.download.PipSession())

requirements = [str(ir.req) for ir in install_reqs if ir is not None]

from consul2nginx import __version__

setup(name             = 'consul2nginx',
      author           = 'Aljosha Friemann',
      author_email     = 'aljosha.friemann@gmail.com',
      description      = 'Create nginx configuration from consul',
      url              = 'https://github.com/afriemann/consul2nginx',
      keywords         = ['consul', 'proxy', 'nginx', 'configuration'],
      classifiers      = [],
      version          = __version__,
      license          = read('LICENSE.txt'),
      long_description = read('README.rst'),
      install_requires = requirements,
      packages         = [ 'consul2nginx' ],
      entry_points     = { 'console_scripts': ['consul2nginx=consul2nginx.cli:main'] },
      package_data     = { 'consul2nginx': ['consul2nginx/templates/*'] },
      platforms        = 'linux',
      include_package_data = True
)

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4 fenc=utf-8
